/*
Created by: X
http://www.createdbyx.com/

This work is licensed under a Creative Commons Attribution-Noncommercial-Share Alike 2.5 Canada License 
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/. 
Permissions beyond the scope of this license may be available at http://contact.createdbyx.com/.
*/
using System.Collections;
using System.Collections.Generic;
using System;


namespace xCollectiveXNA
{
    /// <summary>
    /// A generic index based collection that implements the IList interface.
    /// </summary>
    /// <typeparam name="T">The type that will be stored in the collection.</typeparam>
    public class GenericCollection<T> : IList<T>, IDisposable
    {
        /// <summary>
        /// Provides a delegate for the BeforeAdded event.
        /// </summary>
        /// <param name="sender">A reference to a object that is raising the event.</param>
        /// <param name="Cancel">If set to true the item will not be added.</param>
        /// <param name="Value">A reference to the item that is going to be added.</param>
        public delegate void BeforeAddedEventHandler(object sender, ref bool Cancel, T Value);
        /// <summary>
        /// This event is raised before a item is added to the collection.
        /// </summary>
        public event BeforeAddedEventHandler BeforeAdded;

        /// <summary>
        /// Provides a delegate for the AfterAdded event.
        /// </summary>
        /// <param name="sender">A reference to a object that is raising the event.</param>
        /// <param name="Value">A reference to the item that is was to be added.</param>
        public delegate void AfterAddedEventHandler(object sender, T Value);
        /// <summary>
        /// This event is raised after a item has been added to the collection.
        /// </summary>
        public event AfterAddedEventHandler AfterAdded;

        public delegate void BeforeInsertEventHandler(object sender, ref bool Cancel, int Index, T Value);
        public event BeforeInsertEventHandler BeforeInsert;

        public delegate void AfterInsertEventHandler(object sender, int Index, T Value);
        public event AfterInsertEventHandler AfterInsert;

        //public delegate void BeforeRemoveEventHandler(object sender, ref bool Cancel, T Value);
        public event BeforeRemoveEventHandler BeforeRemove;

        //  public delegate void AfterRemoveEventHandler(object sender, T Value);
        public event AfterRemoveEventHandler AfterRemove;

        public delegate void BeforeRemoveEventHandler(object sender, ref bool Cancel, int Index, T Value);
        public event BeforeRemoveEventHandler BeforeRemoveAt;

        public delegate void AfterRemoveEventHandler(object sender, int Index, T Value);
        public event AfterRemoveEventHandler AfterRemoveAt;

        public delegate void BeforeChangeEventHandler(object sender, ref bool Cancel, int Index, T Value);
        public event BeforeChangeEventHandler BeforeChange;

        public delegate void AfterChangeEventHandler(object sender, int Index, T Value);
        public event AfterChangeEventHandler AfterChange;

        public delegate void BeforeClearEventHandler(object sender, ref bool Cancel);
        public event BeforeClearEventHandler BeforeClear;

        public delegate void AfterClearEventHandler(object sender);
        public event AfterClearEventHandler AfterClear;


        protected internal System.Collections.Generic.List<T> mobjItems = new System.Collections.Generic.List<T>();
        protected internal bool mblnCallDisposeOnItems = false;

        public void Sort()
        {
            this.mobjItems.Sort();
        }

        public void Sort(Comparison<T> comparison)
        {
            this.mobjItems.Sort(comparison);
        }

        public void Sort(IComparer<T> comparer)
        {
            this.mobjItems.Sort(comparer);
        }

        public void Sort(int index, int count, IComparer<T> comparer)
        {
            this.mobjItems.Sort(index, count, comparer);
        }

        public virtual bool CallDisposeOnItems
        {
            get
            {
                return mblnCallDisposeOnItems;
            }
            set
            {
                mblnCallDisposeOnItems = value;
            }
        }

        public virtual void Add(T item)
        {
            if (item == null) return;
            if (BeforeAdded != null)
            {
                bool Cancel = false;
                Delegate[] l;
                l = this.BeforeAdded.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeAddedEventHandler ev = (BeforeAddedEventHandler)l[idx];
                    ev(this, ref Cancel, item);

                    // check if canceled
                    if (Cancel) return;
                }
            }

            // add layer
            mobjItems.Add(item);
            // raise after event
            if (AfterAdded != null) AfterAdded(this, item);
        }

        public virtual void Add(T[] item)
        {
            if (item == null) return;
            int idx = 0;
            while (idx < item.Length)
            {
                this.Add(item[idx]);
                idx++;
            }
        }

        public virtual int Count
        {
            get { return mobjItems.Count; }
        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            return mobjItems.GetEnumerator();
        }

        public virtual IEnumerator GetEnumerator()
        {
            return mobjItems.GetEnumerator();
        }


        public bool RaiseBeforeClear(object sender)
        {
            if (BeforeClear != null)
            {
                bool cancel = false;
                Delegate[] l;
                l = this.BeforeClear.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeClearEventHandler ev = (BeforeClearEventHandler)l[idx];
                    ev(sender, ref cancel);

                    // check if canceled
                    if (cancel) return true;
                }
            }

            return false;
        }

        public void RaiseAfterClear(object sender)
        {
            if (AfterClear != null) AfterClear(sender);
        }

        public virtual void Clear()
        {
           if(this.RaiseBeforeClear(this)) return;
  
            // clear items
            if (mblnCallDisposeOnItems)
            {
                foreach (object i in mobjItems)
                {
                    if (i is IDisposable)
                    {
                        ((IDisposable)i).Dispose();
                    }
                }
            }
            mobjItems.Clear();

            this.RaiseAfterClear(this);
        }

        public virtual bool Remove(T item)
        {
            if (!mobjItems.Contains(item)) return true;
            int index = this.mobjItems.IndexOf(item);
            if (BeforeRemove != null)
            {
                bool Cancel = false;
                Delegate[] l;
                l = this.BeforeRemove.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeRemoveEventHandler ev = (BeforeRemoveEventHandler)l[idx];
                    ev(this, ref Cancel, index, item);

                    // check if canceled
                    if (Cancel) return false;
                }
            }

            // add layer
            mobjItems.Remove(item);
            // raise after event
            if (AfterRemove != null) AfterRemove(this, index, item);

            return true;
        }

        public virtual int IndexOf(T item)
        {
            return mobjItems.IndexOf(item);
        }

        public virtual void Insert(int index, T item)
        {
            if (BeforeInsert != null)
            {
                bool Cancel = false;
                Delegate[] l;
                l = this.BeforeInsert.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeInsertEventHandler ev = (BeforeInsertEventHandler)l[idx];
                    ev(this, ref Cancel, index, item);

                    // check if canceled
                    if (Cancel) return;
                }
            }

            // add layer
            mobjItems.Insert(index, item);
            // raise after event
            if (AfterInsert != null) AfterInsert(this, index, item);
        }

        public virtual T this[int index]
        {
            get
            {
                return mobjItems[index];
            }
            set
            {
                bool Cancel = false;
                if (BeforeChange != null) BeforeChange(this, ref Cancel, index, value);

                // check if canceled
                if (Cancel) return;

                // add layer
                mobjItems[index] = value;
                // raise after event
                if (AfterChange != null) AfterChange(this, index, value);
            }
        }

        public virtual void RemoveAt(int index)
        {
            if (index < 0 || index > mobjItems.Count - 1) return;
            T itemRef = this.mobjItems[index];

            if (BeforeRemoveAt != null)
            {
                bool Cancel = false;
                Delegate[] l;
                l = this.BeforeRemoveAt.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeRemoveEventHandler ev = (BeforeRemoveEventHandler)l[idx];
                    ev(this, ref Cancel, index, itemRef);

                    // check if canceled
                    if (Cancel) return;
                }
            }

            // add layer
            mobjItems.RemoveAt(index);
            // raise after event
            if (AfterRemoveAt != null) AfterRemoveAt(this, index, itemRef);
        }

        public virtual void CopyTo(T[] items, int arrayindex)
        {
            this.mobjItems.CopyTo(items, arrayindex);
        }

        public virtual bool Contains(T item)
        {
            return mobjItems.Contains(item);
        }


        #region " IDisposable Support "
        private bool disposedValue = false; // To detect redundant calls

        // IDisposable
        protected virtual void Dispose(bool disposing)
        {
            if (!this.disposedValue)
            {
                if (disposing)
                {
                    // TODO: free unmanaged resources when explicitly called
                    if (mblnCallDisposeOnItems)
                    {
                        foreach (object i in mobjItems)
                        {
                            if (i is IDisposable)
                            {
                                ((IDisposable)i).Dispose();
                            }
                        }
                    }
                    mobjItems.Clear();
                    mobjItems = null;
                }

                // TODO: free shared unmanaged resources
            }
            this.disposedValue = true;
        }

        // This code added by Visual Basic to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion


    }

}
