/*
Created by: X
http://www.createdbyx.com/

xCollectiveXNA is licenced under the Creative Commons Attribution-NonCommercial-ShareAlike 2.5 Canada
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/
*/
using System.Collections;
using System.Collections.Generic;
using System;

namespace xCollectiveXNA
{
    public class GenericDictionary<TKey, TValue> : IDictionary<TKey, TValue>, IDisposable
    {

        public delegate void BeforeAddedEventHandler(object sender, TKey key, TValue value, ref bool Cancel);
        public event BeforeAddedEventHandler BeforeAdded;

        public delegate void AfterAddedEventHandler(object sender, TKey key, TValue value);
        public event AfterAddedEventHandler AfterAdded;

        public delegate void BeforeRemoveEventHandler(object sender, TKey key, TValue value, ref bool Cancel);
        public event BeforeRemoveEventHandler BeforeRemove;

        public delegate void AfterRemoveEventHandler(object sender, TKey key, TValue value);
        public event AfterRemoveEventHandler AfterRemove;

        public delegate void BeforeChangeEventHandler(object sender, TKey key, TValue value, ref bool Cancel);
        public event BeforeChangeEventHandler BeforeChange;

        public delegate void AfterChangeEventHandler(object sender, TKey key, TValue value);
        public event AfterChangeEventHandler AfterChange;

        public delegate void BeforeClearEventHandler(object sender, ref bool Cancel);
        public event BeforeClearEventHandler BeforeClear;

        public delegate void AfterClearEventHandler(object sender);
        public event AfterClearEventHandler AfterClear;


        protected internal Dictionary<TKey, TValue> mobjItems = new Dictionary<TKey, TValue>();
        protected internal bool mblnCallDisposeOnItems = false;


        public delegate void KeyChangedEventHandler(object sender, TKey oldkey, TKey newkey);
        public event KeyChangedEventHandler KeyChanged;

        public virtual bool ChangeKey(TKey existingkey, TKey newkey)
        {
            // if existing key is not present then exit
            if (!this.ContainsKey(existingkey)) return false;
            // if new key is already present then exit
            if (this.ContainsKey(newkey)) return false;
            // get existing item
            TValue item = mobjItems[existingkey];
            // remove existing item
            mobjItems.Remove(existingkey);
            // add existing item with new key
            mobjItems.Add(newkey, item);
            // raise KeyChanged event 
            if (this.KeyChanged != null) this.KeyChanged(this, existingkey, newkey);
            return true;
        }

        public Dictionary<TKey, TValue>.KeyCollection Keys
        {
            get { return mobjItems.Keys; }
        }

        public Dictionary<TKey, TValue>.ValueCollection Values
        {
            get { return mobjItems.Values; }
        }

        public virtual bool CallDisposeOnItems
        {
            get
            {
                return mblnCallDisposeOnItems;
            }
            set
            {
                mblnCallDisposeOnItems = value;
            }
        }

        public virtual void Add(TKey Key, TValue item)
        {
            bool cancel = false;

            if (BeforeAdded != null)
            {
                bool Cancel = false;
                Delegate[] l;
                l = this.BeforeAdded.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeAddedEventHandler ev = (BeforeAddedEventHandler)l[idx];
                    ev(this, Key, item, ref Cancel);

                    // check if canceled
                    if (Cancel) return;
                }
            }

            // add layer
            mobjItems.Add(Key, item);
            if (AfterAdded != null) AfterAdded(this, Key, item);
        }

        public virtual void Add(KeyValuePair<TKey, TValue> pair)
        {
            this.Add(pair.Key, pair.Value);
        }

        public virtual void Clear()
        {
            if (BeforeClear != null)
            {
                bool cancel = false;
                Delegate[] l;
                l = this.BeforeClear.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeClearEventHandler ev = (BeforeClearEventHandler)l[idx];
                    ev(this, ref cancel);

                    // check if canceled
                    if (cancel) return;
                }
            }

            // clear items
            if (mblnCallDisposeOnItems)
            {
                foreach (object i in mobjItems.Values)
                {
                    if (i is IDisposable)
                    {
                        ((IDisposable)i).Dispose();
                    }
                }
            }
            mobjItems.Clear();
            if (AfterClear != null) AfterClear(this);
        }

        public virtual int Count
        {
            get { return mobjItems.Count; }
        }

        public virtual bool Remove(KeyValuePair<TKey, TValue> pair)
        {
            return this.Remove(pair.Key);
        }

        public virtual bool Remove(TKey Key)
        {
            if (!this.mobjItems.ContainsKey(Key)) return true;
            TValue item = this.mobjItems[Key];
            if (BeforeRemove != null)
            {
                bool cancel = false;
                Delegate[] l;
                l = this.BeforeRemove.GetInvocationList();
                for (int idx = 0; idx < l.Length; idx++)
                {
                    BeforeRemoveEventHandler ev = (BeforeRemoveEventHandler)l[idx];
                    ev(this, Key, item, ref cancel);

                    // check if canceled
                    if (cancel) return false;
                }
            }


            // remove item
            mobjItems.Remove(Key);
            if (AfterRemove != null) AfterRemove(this, Key, item);
            return true;
        }

        public virtual TValue this[TKey Key]
        {
            get { return mobjItems[Key]; }
            set
            {
                if (!this.mobjItems.ContainsKey(Key)) return;
                // setup args
                bool cancel = false;

                if (BeforeChange != null) BeforeChange(this, Key, value, ref cancel);
                if (cancel) return;
                // remove item
                mobjItems[Key] = value;
                if (AfterChange != null) AfterChange(this, Key, value);
            }
        }


        #region " IDisposable Support "
        private bool disposedValue = false; // To detect redundant calls

        // IDisposable
        protected virtual void Dispose(bool disposing)
        {
            if (!this.disposedValue)
            {
                if (disposing)
                {
                    // TODO: free unmanaged resources when explicitly called
                    if (mblnCallDisposeOnItems)
                    {
                        foreach (object i in mobjItems.Values)
                        {
                            if (i is IDisposable)
                            {
                                ((IDisposable)i).Dispose();
                            }
                        }
                    }
                    mobjItems.Clear();
                    mobjItems = null;
                }

                // TODO: free shared unmanaged resources
            }
            this.disposedValue = true;
        }

        // This code added by Visual Basic to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion


        public virtual bool ContainsKey(TKey key)
        {
            return mobjItems.ContainsKey(key);
        }

        public virtual bool TryGetValue(TKey key, out TValue value)
        {
            return mobjItems.TryGetValue(key, out value);
        }

        public virtual bool Contains(KeyValuePair<TKey, TValue> item)
        {
            return mobjItems.ContainsKey(item.Key) || mobjItems.ContainsValue(item.Value);
        }

        [Obsolete("Method is not implemented.")]
        public virtual void CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex)
        {

        }

        public bool IsReadOnly
        {
            get { return false; }
        }

        public IEnumerator<KeyValuePair<TKey, TValue>> GetEnumerator()
        {
            return mobjItems.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return mobjItems.GetEnumerator();
        }


        #region IDictionary<TKey,TValue> Members


        ICollection<TKey> IDictionary<TKey, TValue>.Keys
        {
            get { return mobjItems.Keys; }
        }

        ICollection<TValue> IDictionary<TKey, TValue>.Values
        {
            get { return mobjItems.Values; }
        }

        #endregion
    }

}
